import gi  # isort:skip

gi.require_version("Gdk", "3.0")
gi.require_version("Gtk", "3.0")
from gi.repository import Gdk, Gtk  # isort:skip

from .config import Config


class Overlay:
    STYLE = b"""
#overlay {
  background-color: @theme_bg_color;
  color: @theme_text_color;
}

.group {
  margin: 10px;
}

.heading {
  font-size: 1.2em;
}

.key {
  font-family: monospace;
  font-weight: normal;
  background-color: @insensitive_bg_color;
  border: 1px solid;
  border-color: @insensitive_base_color;
  color: @theme_unfocused_fg_color;

  padding: 2px;
  margin: 2px;
  font-size: .9em;
}
"""

    def __init__(self):
        window = Gtk.Window(title="alluvium - i3 keybindings", name="overlay")
        window.set_position(Gtk.WindowPosition.CENTER_ALWAYS)
        window.set_type_hint(Gtk.WindowType.POPUP)
        window.set_skip_taskbar_hint(True)
        window.set_skip_pager_hint(True)
        window.set_decorated(False)
        window.set_resizable(False)
        window.set_focus_on_map(False)
        window.set_deletable(False)
        window.set_accept_focus(False)
        window.stick()
        screen = window.get_screen()
        visual = screen.get_rgba_visual()
        window.set_visual(visual)

        self.window = window

        style_provider = Gtk.CssProvider()
        style_provider.load_from_data(Overlay.STYLE)
        Gtk.StyleContext.add_provider_for_screen(
            Gdk.Screen.get_default(),
            style_provider,
            Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION,
        )

        self.box = Gtk.FlowBox(orientation=Gtk.Orientation.HORIZONTAL)
        self.box.set_selection_mode(Gtk.SelectionMode.NONE)

        self.window.add(self.box)
        self.window.show_all()

    def render_group(self, name, bindings):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        box.get_style_context().add_class("group")
        heading = Gtk.Label(label=name)
        heading.get_style_context().add_class("heading")
        box.pack_start(heading, False, False, 0)

        for binding in bindings:
            line = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=10)
            line.pack_start(Gtk.Label(label=binding.label), False, False, 0)
            keys = Gtk.Box(orientation=Gtk.Orientation.HORIZONTAL, spacing=5)
            keys.get_style_context().add_class("keys")
            for key in binding.keys:
                key_label = Gtk.Label(label=key)
                key_label.get_style_context().add_class("key")
                keys.pack_start(key_label, False, False, 0)
            line.pack_end(keys, False, False, 0)
            box.pack_start(line, False, False, 0)

        return box

    def show_bindings(self, mode, config: Config):
        self.box.foreach(lambda label: label.destroy())

        # Ensure that the window shrinks when fewer keybindings are shown
        self.window.resize(1, 1)

        if mode == "default":
            for group, bindings in config.groups:
                self.box.add(self.render_group(group, bindings))
        else:
            matching = [bindings for name, bindings in config.modes if name == mode]
            if len(matching) > 0:
                self.box.add(self.render_group(mode, matching[0]))

        self.box.show_all()



import itertools
from collections import namedtuple

import re


class Config:
    Keybinding = namedtuple("Keybinding", ["group", "label", "keys"])
    REMONTOIRE_ANNOTATION = re.compile(
        r"^\s*##([^#/]*)//([^#/]*)//([^#/]*)##", re.MULTILINE
    )

    def __init__(self, groups, modes):
        self.groups = groups
        self.modes = modes

    @classmethod
    def from_data(cls, data):
        def enters_mode(binding):
            label = binding.label
            return label.startswith("Enter ") and label.endswith(" Mode")

        bindings = cls.parse_config(data)
        all_groups = [
            (group, list(bs))
            for group, bs in itertools.groupby(bindings, lambda b: b.group)
        ]
        groups = []
        modes = []
        for group, bs in all_groups:
            if any(enters_mode(b) for b in bs):
                modes.append((group, bs))
            else:
                groups.append((group, bs))
        mode_group = ("Modes", [])
        for mode, bindings in modes:
            enters = [b for b in bindings if enters_mode(b)]
            mode_group[1].extend(enters)
            for b in enters:
                bindings.remove(b)
        groups.append(mode_group)

        return cls(groups, modes)

    @staticmethod
    def parse_keys(spec):
        keys = []
        i = 0
        while i < len(spec):
            if spec[i] == "<":
                closing = spec.find(">", i + 1)
                if closing == -1:
                    raise Exception(f"Unmatched < in keybinding {spec}")
                else:
                    keys.append(spec[(i + 1) : closing])
                    i = closing + 1
            elif spec[i] == " ":
                i += 1
            else:
                next_space = spec.find(" ", i + 1)
                if next_space == -1:
                    keys.append(spec[i:])
                    break
                else:
                    keys.append(spec[i:next_space])
                    i = next_space

        return keys

    @staticmethod
    def parse_config(data):
        return [
            Config.Keybinding(m[0].strip(), m[1].strip(), Config.parse_keys(m[2]))
            for m in Config.REMONTOIRE_ANNOTATION.findall(data)
        ]



import argparse
import atexit
import os
import signal
import sys
import tempfile
from pathlib import Path
from threading import Thread

from i3ipc import Connection, Event

from .config import Config
from .overlay import Overlay

import gi  # isort:skip

gi.require_version("Gtk", "3.0")
from gi.repository import GLib, Gtk  # isort:skip


def listen_for_mode_change(i3, handler):
    def on_mode_change(_, event):
        new_mode = event.change
        GLib.idle_add(handler, new_mode)

    i3.on(Event.MODE, on_mode_change)

    i3.main()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--hide",
        default=False,
        action="store_true",
        help="Stop a running alluvium instance",
    )
    parser.add_argument(
        "--toggle", default=False, action="store_true", help="Toggle alluvium on/off"
    )
    parser.add_argument("-m", "--mode", default="default", help="Initial mode to show")
    parser.add_argument(
        "--quit-on-default",
        default=False,
        action="store_true",
        help="Quit when returning to default mode",
    )
    args = parser.parse_args()

    hide = args.hide
    toggle = args.toggle
    initial_mode = args.mode
    quit_on_default = args.quit_on_default

    pid_file = Path(tempfile.gettempdir()) / "alluvium.pid"

    if hide:
        if pid_file.is_file():
            os.kill(int(pid_file.read_text()), signal.SIGINT)
        sys.exit(0)
    elif toggle:
        if pid_file.is_file():
            os.kill(int(pid_file.read_text()), signal.SIGINT)
            sys.exit(0)
    else:
        # Never run a second instance on top of another one
        if pid_file.is_file():
            sys.exit(0)

    i3 = Connection()
    config_reply = i3.get_config()
    config = Config.from_data(config_reply.config)

    overlay = Overlay()
    overlay.show_bindings(initial_mode, config)

    def quit(*args):
        i3.main_quit()
        Gtk.main_quit()

    def on_mode_change(mode):
        if mode == "default" and quit_on_default:
            quit()
        else:
            overlay.show_bindings(mode, config)

    # Update the overlay when i3 changes its mode
    thread = Thread(
        name="thread-i3",
        target=listen_for_mode_change,
        args=(i3, on_mode_change),
        daemon=False,
    )
    thread.start()

    overlay.window.connect("destroy", quit)
    signal.signal(signal.SIGINT, lambda *args: quit())

    def cleanup_pid_file():
        pid_file.unlink(missing_ok=True)

    atexit.register(cleanup_pid_file)

    pid_file.write_text(str(os.getpid()))
    Gtk.main()



from setuptools import find_packages, setup

with open("README.md", "r") as f:
    long_description = f.read()

setup(
    name="alluvium",
    version="0.1.1",
    author="Marten Lienen",
    author_email="marten.lienen@gmail.com",
    description="Interactive bindings visualizer for i3",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/cqql/alluvium",
    packages=find_packages(),
    include_package_data=True,
    install_requires=["i3ipc", "PyGObject"],
    entry_points={"console_scripts": ["alluvium=alluvium.main:main"]},
)
